/***********************************************************************
 *
 *   This file contains functions for the automatic complexity calculation
 * branche 1
 *************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "Xbascountpriv.h"


/* number of different count structs
   Init to 1 for "garbage" struct */
static int nbCount = 1 ;
/* "garbage" struct */
static Xbas_count NULL_count = { "",0x7fffffff,0,0,0,0,0,0,0 } ;

/* pointer to the "garbage" struct */
static Xbas_count * firstTbCount[1] = { &NULL_count } ;

/* struct pointer array */
static Xbas_count * *tbCount = firstTbCount ;

/* initial number of allocated counter struct = 1 ("garbage" ) */
static int sizePtCount = 1 ;

/* pointer to the current struct  */
Xbas_count *curCount = &NULL_count ;

/* information for a basicop */
typedef struct OP_INFO {
    XB_typ_obs      type;		/* enum ident	*/
    const char *    name ;		/* name string	*/
    int             cost ;		/* cycles cost	*/
} OP_INFO ;

static OP_INFO Ops_Info[XBT_MAX] = {
        {XBT_unknown        , "???"         , 0 }
        ,{ XBT_add          , "add"         , 1 }
        ,{ XBT_sub          , "sub"         , 1 }
        ,{ XBT_abs_s        , "abs_s"       , 1 }
        ,{ XBT_shl          , "shl"         , 1 }
        ,{ XBT_shr          , "shr"         , 1 }
        ,{ XBT_extract_h    , "extract_h"   , 1 }
        ,{ XBT_extract_l    , "extract_l"   , 1 }
        ,{ XBT_mult         , "mult"        , 1 }
        ,{ XBT_L_mult       , "L_mult"      , 1 }
        ,{ XBT_L_mult_X     , "L_mult_X"    , 1 }
        ,{ XBT_negate       , "negate"      , 1 }
        ,{ XBT_round        , "round"       , 1 }
        ,{ XBT_L_mac        , "L_mac"       , 1 }
        ,{ XBT_L_mac_X      , "L_mac_X"     , 1 }
        ,{ XBT_L_msu        , "L_msu"       , 1 }
        ,{ XBT_L_msu_X      , "L_msu_X"     , 1 }
        ,{ XBT_L_macNs      , "L_macNs"     , 1 }
        ,{ XBT_L_msuNs      , "L_msuNs"     , 1 }

    /* ponderation 2 */
        ,{ XBT_L_add        , "L_add"       , 2 }
        ,{ XBT_L_sub        , "L_sub"       , 2 }
        ,{ XBT_L_add_c      , "L_add_c"     , 2 }
        ,{ XBT_L_sub_c      , "L_sub_c"     , 2 }
        ,{ XBT_L_negate     , "L_negate"    , 2 }
        ,{ XBT_L_shl        , "L_shl"       , 2 }
        ,{ XBT_L_shr        , "L_shr"       , 2 }
        ,{ XBT_mult_r       , "mult_r"      , 2 }
        ,{ XBT_shr_r        , "shr_r"       , 2 }
        ,{ XBT_mac_r        , "mac_r"       , 2 }
        ,{ XBT_msu_r        , "msu_r"       , 2 }
        ,{ XBT_L_deposit_h , "L_deposit_h"  , 2 }
        ,{ XBT_L_deposit_l , "L_deposit_l"  , 2 }

    /* ponderation 3 */
        ,{ XBT_L_shr_r      , "L_shr_r"     , 3 }
        ,{ XBT_L_abs        , "L_abs"       , 3 }

    /* ponderation 4 */
        ,{ XBT_L_sat        , "L_sat"       , 4 }
    /* ponderation 15 */
        ,{ XBT_norm_s       , "norm_s"      , 15 }
    /* ponderation 18 */
        ,{ XBT_div_s        , "div_s"       , 18 }
    /* pondearation 30 */
        ,{ XBT_norm_l       , "norm_l"      , 30 }
    /* ponderation 1 */   
        ,{ XBT_DataMove16   , "Move16"      , 1 }
        ,{ XBT_DataMove32   , "Move32"      , 2 }
        ,{ XBT_Logic16      , "Logic16"     , 1 }
        ,{ XBT_Logic32      , "Logic32"     , 2 }
        ,{ XBT_and          , "and"         , 1 }
        ,{ XBT_or           , "or"          , 1 }
        ,{ XBT_xor          , "xor"         , 1 }
        ,{ XBT_L_and        , "L_and"       , 2 }
        ,{ XBT_L_or         , "L_or"        , 2 }
        ,{ XBT_L_xor        , "L_xor"       , 2 }

        ,{ XBT_Test         , "Test"        , 2 }
} ;


/* information to trace saturations, for a line of "C" soure code */
struct XBC_Line {
    XBC_Line *       next ;			/* link		*/
    X_INT32         indFicName ;	/* index of the C source code name incuding the basicop which arise the saturation	*/
    X_INT32         line ;			/* line number for the basicop */
    X_INT32         count ;			/* counter of the saturations for the "C" line */ 
} ;

/* head for free saturation struct list */
static XBC_Line * freeSat = NULL ; 


/* maximum "C" source files */
#define MAX_XB_FILES    1024 
/* number of "C" files */
static int nbNameFile = 0 ;
/* name pointer array for "C" files */
static const char * tbNameFile[MAX_XB_FILES] ;

/* permutation for the "C" file name, to sort them and to allow fast search */
static  int NameFileRank[MAX_XB_FILES] ;

#define NB_SAT_INCR     4096

#define INCR_COUNT      64


/*################################## INTERNAL FUNCTIONS #########################################*/


/* comparison for 2 file name, using string order */
static int X_USER CmpPtNameFile(const void *e1,const void *e2) {
    return strcmp (tbNameFile[*((int *)e1)],tbNameFile[*((int *)e2)]);
}

/*

  Search a filename in the known filenames.
  If the filename is unknown, it is inserted.
  Return the rank in the filename.

*/
static int XBC_FindNameFile(const char *namFile) {
    int * ptFind ;
/* adds temporary at the end of the list */
    NameFileRank[nbNameFile] = nbNameFile ;
    tbNameFile[nbNameFile] = namFile ;
/* search */
    if ( (ptFind = bsearch(NameFileRank+nbNameFile,NameFileRank,nbNameFile,sizeof(NameFileRank[0]),CmpPtNameFile)) != NULL) {
        /* found, return the rank */
		return NameFileRank[(ptFind - NameFileRank)] ;
    }
	/* not found */
    nbNameFile++; /* add definitively */
	/* maintains the permuation array sorted */
    qsort((void *)NameFileRank,nbNameFile,sizeof(NameFileRank[0]),CmpPtNameFile) ;
    return nbNameFile - 1 ;
}


/* comparison for 2 counter struct, using string name order
   !!!!! WARNING : the "garbage" struct naem is equl to "", ensures that it remains the first in the sorted list 
*/

static int X_USER CmpPtCont(const void *e1,const void *e2) {
    return strcmp( ((Xbas_count **)e1)[0]->name ,((Xbas_count **)e2)[0]->name);
}

/*
  trivial hash for a couple file,line
*/
static int XBC_Hash(int indFic, X_INT32 line) {

    return (indFic ^  line ) & (MAX_HASH_SAT -1) ;
}


/*
  Retrieves a saturation counter by its file name index and line number.
  If the counter does not exits it created.
*/

static XBC_Line * XBC_Find_Sat(XBC_Line * * headSat,int indFic, X_INT32 line) {
    int hash = XBC_Hash(indFic,line);
    XBC_Line * curSat = headSat[hash] ;
    while(curSat != NULL) {
        if((curSat->indFicName == indFic) && (curSat->line == line)) {
            return curSat ; /* found */
        }
        curSat = curSat->next ;
    }
    /* not found, creates a new  */
    if(freeSat == NULL) {
        /* no more free counter */
        int i ;
        freeSat  = malloc(NB_SAT_INCR *sizeof(freeSat[0]));
        for(i=0;i<NB_SAT_INCR-1;i++) { /* links the new free saturation counter */
            freeSat[i].next = freeSat + i + 1 ;
        }
        freeSat[NB_SAT_INCR-1].next = NULL ;
    }
	/* inset the new counter */
    curSat = freeSat ;
    freeSat = freeSat->next ;
    curSat->next = headSat[hash] ;
    headSat[hash] = curSat ;
    curSat->count = 0 ;
    curSat->indFicName = indFic ;
    curSat->line = line ;
    return curSat ;

}

/*
  Retrieves a saturation counter by its file name and line number.
  If the counter does not exits it created.
*/

static XBC_Line * XBC_Find_Sat_byName(XBC_Line **headLine,const char *ficName, X_INT32 line) {
    int indFic = XBC_FindNameFile(ficName) ; 
    return XBC_Find_Sat(headLine,indFic,line) ;
}

/*
  Frame transition for a specific Xbas_count.
  - The frame counters are added to the cumulative counters, and are reseted
  - the frame saturation counter is added to the cumulative saturation counter and is reseted.
  - If the sum of cycles for the frame is greater than 0, the frame counter is incremented and
    min max are updated
*/

static void Do_EndFrame(Xbas_count *xbc) {
        int i ;
		X_INT32 cntCur = 0 ;
        for(i=0;i<XBT_MAX;i++) {
			xbc->cntInstTot[i] += xbc->cntInst[i] ;
			cntCur += xbc->cntInst[i] * Ops_Info[i].cost;
            xbc->cntInst[i] = 0;
        }
		xbc->cntSatTot += xbc->cntSat ;
		xbc->cntSat = 0 ;
		if(cntCur > 0) {
			if(cntCur < xbc->cntMin)  xbc->cntMin = cntCur ;
			if(cntCur > xbc->cntMax)  xbc->cntMax = cntCur ;
			xbc->nbFrames ++ ;
		}
}

/*
  Adds a Xbas_count xbcInc to an other xbcSum.
  In fact it is only the frame counters (not the cumulative ones) that are added.
  The consequences are :
  - if frame counting is not used, it is OK because the cumalatives counter are not used and remains null.
  - if frame counting is used , the addition must be called before each Do_EndFrame.
 
*/
static void Do_Add(Xbas_count *xbcSum,Xbas_count *xbcInc) {
    if(xbcInc != NULL) {
        int i ;
        if(xbcSum == NULL) xbcSum = tbCount[0] ;
        xbcSum->cntSat += xbcInc->cntSat ;
        for(i=0;i<XBT_MAX;i++) {
            xbcSum->cntInst[i] +=xbcInc->cntInst[i] ;
        }
    }
}


/*################################## EXTERNAL FUNCTIONS #########################################*/

/*
  Allocation for a new Xbas_count struct.
  The structure is identifed by is name.
  if the structure was already allocated, its pointer is returned, else the
  pointer of the new allocated structre is returned

*/

Xbas_count * XBC_Alloc(const char *name) {
    {
        Xbas_count * * ptXbc ;
        Xbas_count keyBc  ;
        Xbas_count * ptKeyBc = &keyBc ;
        keyBc.name = (char *) name ;
        /* search a Xbas_count with the same name */
		ptXbc = (Xbas_count * *)bsearch((void *) (&ptKeyBc),tbCount,nbCount,sizeof(tbCount[0]),CmpPtCont );
        if(ptXbc != NULL) {
            /* already existing */ 
            return *ptXbc ;
        } else {
            Xbas_count *xbc ;
            if(sizePtCount == nbCount) { /* no more free Xbas_count */
                if(nbCount == 1) { /* first realloction */
					/*add INCR_COUNT free Xbas_count */
                    tbCount = malloc((sizePtCount + INCR_COUNT) *sizeof(tbCount)) ;
                    tbCount[0] = &NULL_count ;
                    sizePtCount += INCR_COUNT ;
                } else {
					/*add INCR_COUNT free Xbas_count */
                    tbCount = realloc(tbCount,(sizePtCount + INCR_COUNT) *sizeof(tbCount)) ;
                    sizePtCount += INCR_COUNT ;
                }
            }
			/* adds a Xbas_count pointer to the used ones */
            tbCount[nbCount++] = xbc = calloc(1,sizeof(Xbas_count)) ;
            xbc->name = malloc(strlen(name)+1) ;
			xbc->cntMin = 0x7fffffff ; /* min to max value */
            strcpy(xbc->name,name);	
            qsort(tbCount,nbCount,sizeof(tbCount[0]),CmpPtCont );   /* sort the Xbas_count pointer array */
            return xbc ;	/* return the pointer */
        }
    }
}


/*
 Allows to retrieve a Xbas_count by its name. 
 It is of course less efficient than using a pointer, but it simplifies interface between different "C" files.
 Return a pointer to the Xbas_count if it exists, otherway a NULL pointer  
*/
Xbas_count * XBC_Get(const char *name) {
    Xbas_count * * ptXbc ;
    Xbas_count keyBc  ;
    Xbas_count * ptKeyBc = &keyBc ;
    keyBc.name = (char *) name ;
    ptXbc = (Xbas_count * *)bsearch((void *) (&ptKeyBc),tbCount,nbCount,sizeof(tbCount[0]),CmpPtCont );
    if(ptXbc != NULL) {
        /* found */ 
        return *ptXbc ;
    } else {
        return NULL ;
    }
}
/*
 Increments a counter of saturation for a specified Xbas_count.
 If a saturation count for the supplied file name and line count does not exist in the Xbas_count,
 it is added.
*/

void XBC_Inc_Sat(Xbas_count * xbc,const char *ficName, X_INT32 line) {
    if((xbc->mode & XBC_COUNT_SAT_OFF) == 0) { /* to allow skipping saturation recording */	
        XBC_Line * sat = XBC_Find_Sat_byName(xbc->headSat,ficName,line) ; /* search and add if necessary the saturation counter */
        xbc->cntSat++ ;	/* increments the total saturation counter for the Xbas_count */
        sat->count++ ;	/* increments the saturation counter por the specific line and file */
    }
}

void XBC_Inc_Count(Xbas_count * xbc,XB_typ_obs instruct,const char *ficName, X_INT32 line) {
	xbc->cntInst[instruct]++ ;
    if((xbc->mode & XBC_COUNT_LINE_ON) != 0) { 
        XBC_Line * lin = XBC_Find_Sat_byName(xbc->headCount,ficName,line) ; /* search and add if necessary the saturation counter */
        if(lin->count == 0) { /* new line */
			xbc->nbLines++ ;
		}
		lin->count+= Ops_Info[instruct].cost;	/* increments the saturation counter por the specific line and file */
    }
}
void XBC_Dec_Count(Xbas_count * xbc,XB_typ_obs instruct,const char *ficName, X_INT32 line) {
	xbc->cntInst[instruct]-- ;
    if((xbc->mode & XBC_COUNT_LINE_ON) != 0) { 
        XBC_Line * lin = XBC_Find_Sat_byName(xbc->headCount,ficName,line) ; /* search and add if necessary the saturation counter */
        lin->count -= Ops_Info[instruct].cost;	/* increments the saturation counter por the specific line and file */
    }
}

/*
 can allow skipping saturation recording if mode != 0 
*/

void XBC_SetMode(Xbas_count *xbc,int mode) {
    xbc->mode = mode ;
}


/*
Changes the current Xbas_count to cumulate wMops, saturation ...
If xbc == NULL, changes the current Xbas_count to the "garbage" one
*/
void XBC_SetCurrent(Xbas_count *xbc ) {
    if(xbc != NULL) {
        curCount = xbc ;
    } else {
        curCount = tbCount[0] ;
    }
}
/*
  Make a frame transition for a specified Xbas_count.
  If the specified Xbas_count pointer is NULL, the frame transition is made for all the existing
  Xbas_count.
  For each concerned Xbas_count.
  - The frame counters are added to the cumulative counters, and are reseted
  - the frame saturation counter is added to the cumulative saturation counter and is reseted.
  - If the sum of cycles for the frame is greater than 0, the frame counter is incremented and
    min max are updated
*/

void XBC_EndFrame(Xbas_count *xbc) {
	if(xbc != NULL) {
		Do_EndFrame(xbc);
	} else {
		int i ;
		for(i=0;i<nbCount;i++) Do_EndFrame(tbCount[i]);
	}
}


/*
  Adds a specified Xbas_count xbcInc to a specified Xbas_count xbcSum.
  If xbcInc is a null pointer, all the existing Xbas_count (including "garbage"), except xbcSum are added to xbcSum.

  For each XBas_count addition, it is only the frame counters (not the cumulative ones) that are added.
  The consequences are :
  - if frame counting is not used, it is OK because the cumalatives counter are not used and remains null.
  - if frame counting is used , the addition must be called before each Do_EndFrame. As Do_endFrame resets the frame counter,
    it is coherent.
*/
void XBC_Add(Xbas_count *xbcSum,Xbas_count *xbcInc) {
	if(xbcInc != NULL) {
		Do_Add(xbcSum,xbcInc);
	} else {
		int i ;
		for(i=0;i<nbCount;i++) {
			if(tbCount[i] != xbcSum) Do_Add(xbcSum,tbCount[i]);
		}
	}
}

/*
 Resets a Xbas_count structure. All counters are set to 0, and saturations are freed.  
*/
void XBC_Clear(Xbas_count *xbc) {
    if(xbc != NULL) {
        int i ;
        xbc->cntSatTot = xbc->cntSat = 0 ;
        for(i=0;i<XBT_MAX;i++) {
            xbc->cntInst[i] = 0;
			xbc->cntInstTot[i] = 0 ;
        }
		xbc->cntMin = 0x7fffffff ;
		xbc->cntMax = 0 ;
        for(i=0;i<MAX_HASH_SAT;i++) {
            /* frees saturation counters */
            while(xbc->headSat[i] != NULL) {
                XBC_Line * nxt = xbc->headSat[i]->next ;
                xbc->headSat[i]->next = freeSat ;
                freeSat = xbc->headSat[i] ;
                xbc->headSat[i] = nxt ;
            }
        }
    }
}

static int X_USER SortLines(const void *el1,const void *el2) {
	return ((XBC_Line **)el2)[0]->count - ((XBC_Line **)el1)[0]->count ;
} ;
/*
 Prints information for a specific Xbas_count xbc.
 if the Xbas_count pointer is NULL, prints the "garbage" Xbas_count.
 Flags in flgsPrt controls how printing is verbose.
*/
static void XBC_DoPrint(Xbas_count *xbc,int flgsPrt,FILE *stdPrnt) {
    if(xbc == NULL) {
        xbc = tbCount[0] ;
    }
    fprintf(stdPrnt,"Counter '%s' :",xbc->name);
    {
        int i; 
        int count = 0 ;
        int nbInst = 0 ;
        int sat = xbc->cntSat + xbc->cntSatTot ;
        for(i=0;i<XBT_MAX;i++) {
			/* compute the global cycle count, and the global basicop count */
            count += (xbc->cntInst[i] + xbc->cntInstTot[i]) * Ops_Info[i].cost ;
            nbInst += ( xbc->cntInst[i] + xbc->cntInstTot[i]);
        }
        fprintf(stdPrnt,"Total %8d cycles for %d instructions %d saturations",count,nbInst,sat) ;
		if(xbc->nbFrames == 0) { /* frame counting ont used */
			fprintf(stdPrnt," No frame details\n");
		} else {
			fprintf(stdPrnt," %d frames [min %d  mean %d max %d ] cycles by frame \n",
				xbc->nbFrames, xbc->cntMin,(count + xbc->nbFrames / 2) /xbc->nbFrames, xbc->cntMax) ;
		}
        if((flgsPrt & XBC_PRT_INST) != 0){	/* information basicop by basicop */
            for(i=0;i<XBT_MAX;i++) {
                if( (xbc->cntInst[i]+xbc->cntInstTot[i]) != 0 ){
                    fprintf(stdPrnt,"       %8d cycles = %d x %s(%d) \n",(xbc->cntInst[i]+xbc->cntInstTot[i])*Ops_Info[i].cost,(xbc->cntInst[i]+xbc->cntInstTot[i]),Ops_Info[i].name,Ops_Info[i].cost) ;
                }
                count += xbc->cntInst[i] * Ops_Info[i].cost ;
            }
        }
        if((sat > 0 ) && ((flgsPrt & XBC_PRT_SAT) != 0)) { /* detailed saturation informations */
            for(i=0;i<MAX_HASH_SAT;i++) {
                XBC_Line * curSat = xbc->headSat[i] ;
                while(curSat != NULL) {
                    fprintf(stdPrnt," %3d saturations line %4d \"%s\"\n",curSat->count,curSat->line,tbNameFile[curSat->indFicName]);
                    curSat = curSat->next;
                }
            }
        }
        if((flgsPrt & XBC_PRT_LINES) != 0) { /* detailed saturation informations */
			XBC_Line ** tbLines = calloc(sizeof(tbLines[0]),xbc->nbLines) ;
            X_INT32 nbFoundLines = 0 ;
			for(i=0;i<MAX_HASH_SAT;i++) {
                XBC_Line * curLin = xbc->headCount[i] ;
                while(curLin != NULL) {
					if(nbFoundLines < xbc->nbLines) {
						tbLines[nbFoundLines++] = curLin ;
					} else {
						fprintf(stdPrnt," Internal error on line counts \n");
					}
                    curLin = curLin->next;
                }
            }
			/* sorts on cycles number */
			qsort(tbLines,nbFoundLines,sizeof(tbLines[0]),SortLines);
            for(i=0;i<nbFoundLines;i++) {
                XBC_Line * curLin = tbLines[i] ;
                fprintf(stdPrnt," %7d cycles line %4d \"%s\"\n",curLin->count,curLin->line,tbNameFile[curLin->indFicName]);
            }
			free(tbLines);
        }
		
    }
}


void XBC_Print(Xbas_count *xbc,int flgsPrt,FILE *stdPrnt) {
    if(xbc == NULL) {
		int i ;
		for(i=nbCount-1;i>1;i--) {
	  	    XBC_DoPrint(tbCount[i] ,flgsPrt,stdPrnt);
		}
	  	XBC_DoPrint(tbCount[0] ,flgsPrt,stdPrnt);
		if(nbCount>1) {
		  	XBC_DoPrint(tbCount[1] ,flgsPrt,stdPrnt);
		}
    } else {
		XBC_DoPrint(xbc,flgsPrt,stdPrnt);
	}
}
