/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/






/*-----------------
 * File Inclusions
 *-----------------*/
#include <stdlib.h>
#include <stdio.h>

#include "ParmInterface_B.h"
#include "WaveProcExports_B.h"

#include "typedef.h"
#include "basic_op.h"
#include "count.h"

#define WP_EPS      0.2
#define PERCENT_WIDTHPAR    80

struct WaveProcStructX
{
  X_INT16    *TeagerFilter16;
  X_INT32     *TeagerWindow32;
  int         TeagerOnset;
  int         FrameLength;
};
    X_INT16 expFact ;


/*------------
 * Prototypes
 *------------*/
static void TeagerEng (WaveProcStructX *WPX,X_INT32 *Data,X_INT16 expFact);
static void GetTeagerFilter (WaveProcStructX *WPX,X_INT16 *MaxPos);
static int  GetMaximaPositions (unsigned X_INT32 *teager,X_INT16 *MaxPos,X_INT16 Length);

/*-----------
 * Functions
 *-----------*/

/*----------------------------------------------------------------------------
 * FUNCTION NAME: GetMaximaPositions
 *
 * PURPOSE:  Find Max positions      
 *
 * INPUT: 
 *     - teager : teager array where to find max
 *     - Length : size of teager array
 *
 * OUTPUT:
 *     - MaxPos : found Max positions
 *
 * RETURN VALUE: number of max positions (0<= < 10)
 *   
 *
 *---------------------------------------------------------------------------*/
static int GetMaximaPositions (
    unsigned X_INT32 *teager,   /* (i) teager array where to find max */
    X_INT16 *MaxPos,            /* (o) found Max positions           */
    X_INT16 Length              /* (i) size of teager                */
)
{
    X_INT16 i,counterR,counterL ; 
    unsigned int    teager_max;
    X_INT16   maxFound;
    X_INT16 * maxR  ;
    X_INT16 maxL[10] ;
    maxL[0] = 0 ; move16();  
    maxFound =0; move16();
    teager_max = 0 ; move16();

    for (i=0 ; i<Length ; i++){
        test();
        if(L_sub(teager[i],teager_max) > 0) {
            maxFound = 1;  move16();
            teager_max = teager[i]; move32();
            maxL[0] = i; move16();
        }
    }
    counterL = 0;  move16();
    test();  
    if (maxFound != 0){
        X_INT16 L25 = sub(Length,25);
    /* Search MaxL */
        test(); test();
        while((sub(maxL[counterL],25)> 0) && (maxFound != 0)){
            X_INT16 imin = sub(maxL[counterL],80) ;
            X_INT16 isup = sub(maxL[counterL],25) ;
            test();
            if(imin < 0) {
                imin = -1 ;  move16();
                test();
                if(isup < 0) {
                    isup = 0 ;  move16();
                }
            }
            teager_max = L_mult(0,0) ;
            maxFound = 0;  move16();
            for (i=isup;i>imin;i--) {
                test();
                if(L_sub(teager[i],teager_max) >= 0) {
                    maxFound = 1; move16();
                    teager_max = teager[i]; move32();
                    maxL[add(counterL,1)] = i; move16();
                }
            }
            counterL = add(counterL,maxFound) ;
            test(); test();
        }
        for (i=0 ; i<=counterL ; i++) {
            MaxPos[counterL-i] = maxL[i]; move16();
        }

    /* Search MaxR */
        maxR = MaxPos + counterL ;
        counterR = 0;  move16();
        test(); test();
        while ((sub(maxR[counterR],L25) < 0 ) && (maxFound != 0)){
            X_INT16 isup = add(maxR[counterR],80) ;
            X_INT16 imin = add(maxR[counterR],25) ;
            test();
            if(sub(Length,isup) < 0) {
                isup = Length; move16();
                test();
                if(sub(Length,imin) < 0) {
                    imin = Length ; move16();
                }
            }
            teager_max = L_mult(0,0) ; 
            maxFound = 0;  move16();
            for (i=imin;i<isup;i++) {
                test();
                if(L_sub(teager[i],teager_max) >= 0) {
                    maxFound = 1; move16();
                    teager_max = teager[i]; move32();
                    maxR[add(counterR,1)] = i; move16();
                }
            }
            counterR = add(counterR,maxFound) ;             
            test(); test();
        }
        return add(counterR,add(counterL,1));
    }   else {
        return (0);
    }
}


 


/*----------------------------------------------------------------------------
 * FUNCTION NAME: TeagerEng
 *
 * PURPOSE:  Compute Teager energy from input data (time domain)
 *           engT[t] = abs(data[t]*data[t]-data[t-1]*data[t+1])     
 *           if expFact > 0 TeagerWindow32[t] = abs(Data32[t]*Data32[t]-Data32[t-1]*Data32[t+1]) / 2**(2*expFact) 
 *           if expFact ==0 TeagerWindow32[t] = abs(Data32[t]*Data32[t]-Data32[t-1]*Data32[t+1]) / 2
 * INPUT: 
 *     - WPX    :    wave processing context
 *     - Data32 :    Input Data
 *     - expFact :   shift factor for Data32 (Data32[t] = data[t] << expFact)
 *
 * OUTPUT:
 *     - WPX->TeagerWindow32 : Teager energy
 *
 * RETURN VALUE: none
 *   
 *
 *---------------------------------------------------------------------------*/

static void TeagerEng (
        WaveProcStructX *WPX    /* (i/o) associated structure => TeagerWindow32 */
        , X_INT32 *Data32       /* (i)   data in */
        ,X_INT16 expFact        /* (i)   shift factor for Data32 */
    )
{
    X_INT32 *TeagerWindow32 = WPX->TeagerWindow32;
    int i;
    X_INT16 dec = add(1,shl(expFact,1)) ;
    test(); sub(dec,1);
    if(dec<=1)  /* add to avoid saturation in next module */
     {
       dec = 2; move16(); 
     }

    {
        X_INT16 ah0 ,al0, ah1, al1, ah2 , al2 ;
        X_INT32 sum ;
        ah0 = extract_h(Data32[0]);
        al0 = 0x7fff & shr(extract_l(Data32[0]),1) ; logic16();
        ah1 = extract_h(Data32[1]);
        al1 = 0x7fff & shr(extract_l(Data32[1]),1) ; logic16();
        
        sum = L_mult(2,mult(ah0,al0));
        sum  = L_mac(sum,ah0,ah0);
               /* it is preferred not to access directly the 
                  Carry and Overflow flags */
		L_macNs(MAX_32, MIN_16, MIN_16);Carry=1 ;
		sum = L_msuNs(sum,1,add(mult(ah0,al1),mult(ah1,al0)));
		L_macNs(MAX_32, MIN_16, MIN_16);Carry=1 ;
        sum = L_msuNs(sum,ah0,ah1);

        TeagerWindow32[0] = L_shr_r(L_abs(sum),dec ) ; move32();
        /* warning : !!! it is assumed that FRAME_LENGTH - 2 is a multiple of 3 */
        for (i=1 ; i<FRAME_LENGTH-1 ;) {
            ah2 = extract_h(Data32[i+1]);
            al2 = 0x7fff & shr(extract_l(Data32[i+1]),1) ; logic16();
            sum = L_mult(2,mult(ah1,al1));
            sum  = L_mac(sum,ah1,ah1);
			L_macNs(MAX_32, MIN_16, MIN_16);Carry=1 ;
            sum = L_msuNs(sum,1,add(mult(ah0,al2),mult(ah2,al0)));
			L_macNs(MAX_32, MIN_16, MIN_16);Carry=1 ;
            sum = L_msuNs(sum,ah0,ah2);
            TeagerWindow32[i++] = L_shr_r(L_abs(sum),dec ) ; move32();

            ah0 = extract_h(Data32[i+1]);
            al0 = 0x7fff & shr(extract_l(Data32[i+1]),1) ; logic16();
            sum = L_mult(2,mult(ah2,al2));
            sum  = L_mac(sum,ah2,ah2);
			L_macNs(MAX_32, MIN_16, MIN_16);Carry=1 ;
            sum = L_msuNs(sum,1,add(mult(ah1,al0),mult(ah0,al1)));
			L_macNs(MAX_32, MIN_16, MIN_16);Carry=1 ;
            sum = L_msuNs(sum,ah1,ah0);
            TeagerWindow32[i++] = L_shr_r(L_abs(sum),dec ) ; move32();

            ah1 = extract_h(Data32[i+1]);
            al1 = 0x7fff & shr(extract_l(Data32[i+1]),1) ; logic16();
            sum = L_mult(2,mult(ah0,al0));
            sum  = L_mac(sum,ah0,ah0);
			L_macNs(MAX_32, MIN_16, MIN_16);Carry=1 ;
            sum = L_msuNs(sum,1,add(mult(ah2,al1),mult(ah1,al2)));
			L_macNs(MAX_32, MIN_16, MIN_16);Carry=1 ;
            sum = L_msuNs(sum,ah2,ah1);
            TeagerWindow32[i++] = L_shr_r(L_abs(sum),dec ) ; move32();
        
        }
        sum = L_mult(2,mult(ah1,al1));
        sum  = L_mac(sum,ah1,ah1);
		L_macNs(MAX_32, MIN_16, MIN_16);Carry=1 ;
        sum = L_msuNs(sum,1,add(mult(ah0,al1),mult(ah1,al0)));
		L_macNs(MAX_32, MIN_16, MIN_16);Carry=1 ;
        sum = L_msuNs(sum,ah0,ah1);
        TeagerWindow32[i] = L_shr_r(L_abs(sum),dec ) ; move32();


    }
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: GetTeagerFilter
 *
 * PURPOSE:  Construct Teager energy based filter
 *           Can take two values :
 *               - HIGH_EPS = 2**15 * WP_EPS / 2
 *               - LOW_EPS =  2**15 * -WP_EPS / 2
 *
 * INPUT: 
 *     - WPX    :    wave processing context (WPX->TeagerWindow32 : Teager energy)
 *     - MaxPos :    Max positions
 *
 * OUTPUT:
 *     - WPX->TeagerFilter16 : Teager energy based filter
 *
 * RETURN VALUE: none
 *   
 *              
 *
 *---------------------------------------------------------------------------*/
static void GetTeagerFilter (WaveProcStructX *WPX,X_INT16 *MaxPos)
{
    int i, j;
    int NumOfMaxima;
    int FrameLength = WPX->FrameLength;
    X_INT32  *TeagerWindow32 = WPX->TeagerWindow32;
    X_INT16  *TeagerFilter16 = WPX->TeagerFilter16;
    /*----------------------------
    * smoothing of Teager energy
    *----------------------------*/
    {   
        X_INT16 k;
        X_INT32 ant;

        k = 0; move16();
        ant = 0; move32();

        for(i=0; i <= 3; i++) {
            TeagerWindow32[-i-1] = TeagerWindow32[0] ; move32();
            ant = L_add(ant,L_add(TeagerWindow32[i],TeagerWindow32[0]))  ;
        }
        sub(FrameLength,4); /* I can not check if FrameLength can be determined at compile time */
        for (i=0; i < FrameLength-4; i++) {
            X_INT32 tmp ;
              tmp = L_add(ant,TeagerWindow32[i+4]) ;
              ant = L_sub(tmp,TeagerWindow32[i-4]) ;
              TeagerWindow32[i-4] = tmp ; move32() ;
        }
        for (; i < FrameLength; i++) {
            X_INT32 tmp ;
            tmp = L_add(ant,TeagerWindow32[FrameLength-1]) ;
            ant = L_sub(tmp,TeagerWindow32[i-4]) ;
            TeagerWindow32[i-4] = tmp ; move32() ;
        }
        NumOfMaxima = GetMaximaPositions (TeagerWindow32-4, MaxPos, FrameLength);
    }
#define LOW_EPS     -3277   /* (- WP_EPS) / 2.0 * 32768.0 */
#define HIGH_EPS     3277   /* (+ WP_EPS) / 2.0 * 32768.0 */
    for (i=0 ; i<FrameLength ; i++) {
        TeagerFilter16[i] = LOW_EPS ; move16();
    }
    test();
    sub(NumOfMaxima,1);
    if (NumOfMaxima > 1)
    {
        sub(NumOfMaxima,1);
        for (i=0 ; i<(NumOfMaxima-1) ; i++){
            X_INT16 jmin = sub(MaxPos[i],4);
            X_INT16 jmax = add( jmin , extract_h(L_mac(0xffff,sub(MaxPos[i+1],MaxPos[i]),0x6666))) ;
            test();
            if(jmin < 0 ) {
                jmin = 0 ;         move16();
                test();
                if(jmax < 0) {
                    jmax = 0 ; move16();
                }
            }
            for(j=jmin;j<jmax;j++) {
                TeagerFilter16[j] = HIGH_EPS ; move16();
            }
        }
        {
            X_INT16 jmin = sub(MaxPos[NumOfMaxima-1],4);
            X_INT16 jmax = add( jmin , extract_h(L_mac(0xffff,sub(MaxPos[NumOfMaxima-1],MaxPos[NumOfMaxima-2]),0x6666))) ;
            test();
            sub(jmax,FrameLength);
            if(jmax > FrameLength ) {
                jmax = FrameLength ; move16();
            }
            for(j=jmin;j<jmax;j++) {
                TeagerFilter16[j] = HIGH_EPS ; move16();
            }
        }
    }
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoWaveProcAlloc_B
 *
 * PURPOSE:  Allocation of a Waveprocessing struct
 *           Warning : all the fields are not allocated (needs a call to DoWaveProcInit_B)
 *
 * INPUT:  N/A
 *
 * OUTPUT: N/A
 *
 * RETURN VALUE: pointer to the new allocated struct
 *   
 *              
 *
 *---------------------------------------------------------------------------*/

extern WaveProcStructX *DoWaveProcAlloc_B (void)
{
  WaveProcStructX *WPX = calloc (1, sizeof (WaveProcStructX));
  return WPX;
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoWaveProcInit_B
 *
 * PURPOSE:  Initialisation of a Waveprocessing struct
 *           Warning : allocations are done in DoWaveProcInit_B,so consecutive calls
 *                     to DoWaveProcInit_B leads to memory leaks.
 *
 * INPUT:
 *    - This : global FEParamsX_F context (for this->WPX, This->FrameLength)
 *
 * OUTPUT:
 *    - initialized this->WPX
 *
 * RETURN VALUE: none
 *   
 *              
 *
 *---------------------------------------------------------------------------*/
extern void DoWaveProcInit_B (FEParamsX_F *This)
{
    WaveProcStructX *WPX = This->WPX;
    
    WPX->FrameLength = This->FrameLength;
    WPX->TeagerOnset = 0;
}


/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoWaveProcDelete_B
 *
 * PURPOSE:  frees a  Waveprocessing struct
 *
 * INPUT:
 *    - WPX : Waveprocessing struct
 *
 * OUTPUT:
 *
 * RETURN VALUE: none
 *   
 *              
 *
 *---------------------------------------------------------------------------*/
extern void DoWaveProcDelete_B (WaveProcStructX *WPX)
{
  if (WPX != NULL)
    free (WPX);
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoWaveProc
 *
 * PURPOSE: Do WaveProcessing for a frame 
 *              
 *
 * INPUT:
 *      - DataInSh32 : Frame input signal   
 *      - shiftIn  :   shift factor for input signal DataInSh32(i] = DataIn[i] * 2**(15+shiftIn)
 *      - This  : global FEParamsX_F context (for this->WPX)                
 *
 * OUTPUT:
 *      - DataOut16 : output processed signal 
        - shiftOut :  shift factor for input signal DataOut16[i] = DataOut[i] * 2**shiftOut
 *
 * RETURN VALUE: TRUE
 *   
 *
 *---------------------------------------------------------------------------*/
extern BOOLEAN  DoWaveProc_B (
        X_INT32 *DataInSh32,    /* (i) Frame input signal (INT32)       */
        X_INT16 shiftIn,        /* (i) shift factor for input signal    */
        X_INT16 *DataOut16,     /* (o) output processed signal (INT16)  */
        X_INT16 *shiftOut,      /* (o) shift factor for output signal   */
        FEParamsX_F *This       /* (i) associated structure             */
)
{
    WaveProcStructX *WPX = This->WPX;
    int i;
    X_INT16 doSWP;
    X_INT32 TeagerWindow32[FRAME_LENGTH+4]; /* +4 to reserve place for smoothing */
    X_INT16 MaxPos[10] ;
    X_INT32 Energy32 ;
    X_INT32 *DataIn32 = DataInSh32 ;



     doSWP = 1; move16();

    /*------------------
    * low energy check
    *------------------*/
    {
        X_INT16 sh8 = sub(shiftIn,8);
        Energy32 = - 200 * 65536 ; move32();

        for (i=0; i<WPX->FrameLength; i+=4) {
            X_INT16 sig ;
            sig = extract_l(L_shr_r(DataInSh32[i],sh8));
            Energy32 = L_mac(Energy32,sig,sig) ;
            sig = extract_l(L_shr_r(DataInSh32[i+1],sh8));
            Energy32 = L_mac(Energy32,sig,sig) ;
            sig = extract_l(L_shr_r(DataInSh32[i+2],sh8));
            Energy32 = L_mac(Energy32,sig,sig) ;
            sig = extract_l(L_shr_r(DataInSh32[i+3],sh8));
            Energy32 = L_mac(Energy32,sig,sig) ;
            test();
            if(Energy32 >= 0) break ;
        }
        test();
        if (Energy32 < 0) doSWP = 0;
        expFact = sub(shiftIn,15) ;
    }
   test();    
   if (doSWP) 
    {  
       WPX->TeagerWindow32 = TeagerWindow32+4 ; /* to keep place for smoothing */
       WPX->TeagerFilter16 = (X_INT16 *)TeagerWindow32 ;
        /*-----------------------
        * compute Teager energy
        *-----------------------*/
        TeagerEng (WPX, DataIn32,expFact);

        /*-------------------
        * get Teager filter
        *-------------------*/
        GetTeagerFilter (WPX, MaxPos);
        { 
            int i ; 
            /*---------------------
            * apply Teager filter
            *---------------------*/
            X_INT16  Win16 ;

            DataOut16[-1] =round(DataIn32[-1]); move16();
            sub(WPX->FrameLength,1);
            for (i=0 ; i<WPX->FrameLength-1 ; i++){
                X_INT16 hi = extract_h(DataIn32[i]);
                X_INT16 lo = extract_l(DataIn32[i]);
                hi = sub(hi,shr(lo,15)) ;
                Win16 = add (WPX->TeagerFilter16[i],WPX->TeagerFilter16[i+1]);
                DataOut16[i] = mac_r(L_mac(DataIn32[i],1,mult_r(lo,shr(Win16,1))),Win16,hi) ; move16();
            }
            {   /* last sample */
                X_INT16 hi = extract_h(DataIn32[i]);
                X_INT16 lo = extract_l(DataIn32[i]);
                hi = sub(hi,shr(lo,15)) ;
                Win16 = add (WPX->TeagerFilter16[i],WPX->TeagerFilter16[i]);
                DataOut16[i] = mac_r(L_mac(DataIn32[i],1,mult_r(lo,shr(Win16,1))),Win16,hi) ; move16();
            }
            *shiftOut = sub(expFact,1)  ;
        }
       WPX->TeagerWindow32 = NULL ; /* not necessary but avoids  dangling pointers */ 
       WPX->TeagerFilter16 = NULL ;
   
   } else {
        for(i=-1;i<WPX->FrameLength;i++) {
            DataOut16[i] = round(DataIn32[i]); move16();
        }
        *shiftOut = sub(expFact,1)  ;
    }
    return TRUE;
}
