/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/
/*-------------------------------------------------------------------------------
 *
 * FILE NAME: decoder.h
 * PURPOSE:   header for decoder of DSR compression
 *
 *-------------------------------------------------------------------------------*/

/* Last speech frame detection for frame detected with bad crc */
/*if in a crc errored frame, if ZEROS_THRESHOLD out of 7 are zero, assume that this is the lastframe + 1*/
#define ZEROS_THRESHOLD 5 
/* Search for errored frames missed by CRC */
/* When looking at thresholds if ERRORS_THRESHOLD or greater are over the thres then decide it is in error. */
#define ERRORS_THRESHOLD 2 
/*in relation to s.d.'s of differences from training data*/
#define DETECTION_THRES_COEF 3.0



/*Structures to hold the bitstream.*/

typedef struct dframe{
   unsigned char a[7];
   unsigned char b[7];
   unsigned char crc;
#ifdef XAFE
   unsigned char pitch1_index;
   unsigned char pitch2_index;
   unsigned char class1_index;
   unsigned char class2_index;
#endif
}dframe;


typedef struct mframe{
   unsigned char sync[2];
   unsigned char golay[4];
#ifdef XAFE
   unsigned char data[162];
#else
  unsigned char data[138];
#endif
}mframe;


typedef struct golaydata{
  unsigned short int data;
  unsigned short int crc;
}golaydata;





int main(int argc, char *argv[]);
void init_thresholds(void);
int process_two_frames(dframe **BufferedDFrames, dframe *CurrentDFrame, dframe *LastGoodDFrame, dframe *PreviousDFrame,  
                        int *dataframes_buffered, int endFrames, int buffering_data_mode);
int find_good_golay(mframe **BufferedMultiFrames, golaydata **BufferedGolays, short int *GoodGolay, int *frames_buffered);
void add_to_buffer(dframe **BufferedData, dframe *ToBuffer, int buffered_count);
int threshold_error(dframe DataFrame);
void dequantise_dataframe(HTK_DataFrame *result, unsigned char given_index[7]);
#ifdef XAFE
void error_correct_and_write(dframe *Left, int Left_a_or_b, dframe *Right, int Right_a_or_b, dframe *BufferedData, int count);
#else
void error_correct_and_write(unsigned char Left[7], unsigned char Right[7], dframe *BufferedData, int count);
#endif
void get_next_golay(golaydata **BufferedGolays, mframe BufferedMultiFrames, int golays_buffered);
int get_next_frame(mframe **BufferedMultiFrames, int frames_buffered);
int sync_error(unsigned char *given_sync, unsigned short int sync);
int get_two_data_frames(dframe *DataFrame, mframe BufferedMultiFrame, int frame);
#ifdef XAFE
void write_two_data_frames(dframe given, int numFrames, unsigned char error_status);
#else
void write_two_data_frames(dframe given, int numFrames);
#endif
void process_params(int argc, char *argv[]);
void print_usage(void);
void write_htk_coefs(HTK_DataFrame *given);
void write_htk_header(HTK_Header *header);
void print_htk_coefs(HTK_DataFrame given);
#ifdef XAFE
void inv_quant_pitch_abs(UWord8 cQPIndex, float *pfQuantPeriod);
void inv_quant_pitch_diff(UWord8 cQPIndex, float *pfQPHistory,
                          int *piReliableFlag, float *pfQuantPeriod);
void get_class(UWord8 class_index, float pitch_period, UWord8 *class);
#endif
