/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/
/*-------------------------------------------------------------------------------
 *
 * FILE NAME: 
 *
 *    golay.c
 *
 * PURPOSE:  
 *
 *    Golay and CRC routines for (Extended) DSR Advanced Front-end quantisers.
 *
 *-------------------------------------------------------------------------------*/

#include "typedef.h"
#include "basic_op.h"
#include "count.h"
#include "golay.h"



/*----------------------------------------------------------------------------
 * FUNCTION NAME:
 *
 *    golay_encode
 *
 * PURPOSE:
 *
 *    
 *
 * INPUT:
 *
 *
 * OUTPUT:
 *
 *
 * RETURN VALUE:
 *
 *
 *---------------------------------------------------------------------------*/


unsigned short golay_encode(UWord16 data)
{
                 
  const Word16 g_matrix[16] = {0x808b,
                               0xc0ce,
                               0xe0ed,
                               0x7077,
                               0xb8b0,
                               0x5c58,
                               0x2e2c,
                               0x1716,
                               0x8b01,
                               0x4581,
                               0x22c1,
                               0x1161,
                               0x08b1,
                               0x0459,
                               0x022d,
                               0x0117};
                 
  Word16 bit;
  Word16 mask;
  Word16 crc=0x0;
  
  for (bit=0,mask=0x1; bit<16; bit++,mask<<=1)
  {
     if (mask & data)
     {
        crc ^= g_matrix[bit];
     }
  }

  return crc;
}


/*----------------------------------------------------------------------------
 * FUNCTION NAME:
 *
 *    golay_errors
 *
 * PURPOSE:
 *
 *    
 *
 * INPUT:
 *
 *
 * OUTPUT:
 *
 *
 * RETURN VALUE:
 *
 *
 *---------------------------------------------------------------------------*/



Flag golay_errors(UWord16 *data, UWord16 *crc)
{

   unsigned short data_copy,
      crc_copy;
  
   int i, j, k;

  
   /*everything is ok*/
   if ((*crc) == golay_encode(*data)) return FALSE;



   /*May be 1 bit error in data or crc */


   for(i=0; i<DATA_LEN; i++)
   {
      data_copy = (0x1 << i) ^ (*data);
    
      if((*crc) == golay_encode(data_copy))
      {
         *data = data_copy;
         return FALSE;
      }
   }


   for(i=0; i<DATA_LEN; i++)
   {
      crc_copy = (0x1 << i) ^ (*crc);
    
      if(crc_copy == golay_encode(*data))
      {
         *crc = crc_copy;
         return FALSE;
      }
   }


   /*May be 2 bit errors in either crc/data or both*/
   /*1st check data for 2 errors, 2nd check crc for 2 errors, 3rd, check both for 1 error*/

   for(i=0; i<(DATA_LEN-1); i++)
      for(j=(i+1); j<DATA_LEN; j++)
      {
         data_copy = ((0x1 << i) | (0x1 << j)) ^ (*data);

         if((*crc) == golay_encode(data_copy))
         {
            *data = data_copy;
            return FALSE;
         }
      }



   for(i=0; i<(DATA_LEN-1); i++)
      for(j=(i+1); j<DATA_LEN; j++)
      {
         crc_copy = ((0x1 << i) | (0x1 << j)) ^ (*crc);

         if(crc_copy == golay_encode(*data))
         {
            *crc = crc_copy;
            return FALSE;
         }
      }


   for(i=0; i<DATA_LEN; i++)
   {

      data_copy = (0x1 << i) ^ (*data);

      for(j=0; j<DATA_LEN; j++)
      {
         crc_copy = (0x1 << j) ^ (*crc);

         if(crc_copy == golay_encode(data_copy))
         {
            *crc = crc_copy;
            *data = data_copy;
            return FALSE;
         }
      }
   }



   /*Maybe 3 bit errors*/

   /*Maybe 3 in data or 3 in crc, or 2 in data 1 in crc, or 1 in data 2 in crc*/
  
   for(i=0; i<(DATA_LEN-2); i++)
      for(j=(i+1); j<(DATA_LEN-1); j++)
         for(k=(j+1); k<DATA_LEN; k++)
         {
            data_copy = ((0x1 << i) | (0x1 << j) | (0x1 << k)) ^ (*data);

            if((*crc) == golay_encode(data_copy))
            {
               *data = data_copy;
               return FALSE;
            }
         }


   for(i=0; i<(DATA_LEN-2); i++)
      for(j=(i+1); j<(DATA_LEN-1); j++)
         for(k=(j+1); k<DATA_LEN; k++)
         {
            crc_copy = ((0x1 << i) | (0x1 << j) | (0x1 << k)) ^ (*crc);

            if(crc_copy == golay_encode(*data))
            {
               *crc = crc_copy;
               return FALSE;
            }
         }





   for(i=0; i<(DATA_LEN-1); i++)
      for(j=(i+1); j<DATA_LEN; j++)
      {
         data_copy = ((0x1 << i) | (0x1 << j)) ^ (*data);
      
         for(k=0; k<DATA_LEN; k++)
         {
            crc_copy = (0x1 << k) ^ (*crc);

            if(crc_copy == golay_encode(data_copy))
            {
               *crc = crc_copy;
               *data = data_copy;
               return FALSE;
            }
         }
      }




   for(i=0; i<(DATA_LEN-1); i++)
      for(j=(i+1); j<DATA_LEN; j++)
      {
         crc_copy = ((0x1 << i) | (0x1 << j)) ^ (*crc);
      
         for(k=0; k<DATA_LEN; k++)
         {
            data_copy = (0x1 << k) ^ (*data);

            if(crc_copy == golay_encode(data_copy))
            {
               *crc = crc_copy;
               *data = data_copy;
               return FALSE;
            }
         }
      }


   return TRUE;
}




/*----------------------------------------------------------------------------
 * FUNCTION NAME:
 *
 *    mfcc_crc_encode
 *
 * PURPOSE:
 *
 *    CRC computation for MFCC bits.
 *
 * INPUT:
 *
 *    data[] - VQ indices.
 *
 * OUTPUT:
 *
 *    None.
 *
 * RETURN VALUE:
 *
 *   CRC for input data.
 *
 *---------------------------------------------------------------------------*/



UWord8 mfcc_crc_encode(UWord8 data[])
{
   Word16 i,j;
   Word16 mask = 0xc;  /* equivalent to 1 + x + X**4 */

   Word16 crc;
   Word16 nibble;


   /* lookup table generation --- need not to count complexity */
   static int first = TRUE;
   static unsigned char crc4_lut[16];
   if (first)
   {
      for (i=0; i<16; i++)
      {
         crc = i;
         for (j=0; j<4; j++)
         {
            crc = (crc & 0x1) ? ((crc >> 1) ^ mask) : (crc >> 1);
         }
         crc4_lut[i] = crc;
      }
      first = FALSE;
   }
  

   crc = 0; move16();
   for (i=0; i<22; i++)
   {
      test(); logic16();
      if ( i & 1 )
      {
         nibble = shr((0xf0 & data[ i/2 ]), 4); logic16();
      }
      else
      {
         nibble = 0xf & data[ i/2 ] ; logic16(); 
      }        
     
      crc = crc4_lut[crc ^ nibble]; move16(); logic16();

   }

  
   return crc;
}



/*----------------------------------------------------------------------------
 * FUNCTION NAME:
 *
 *    mfcc_crc_errors
 *
 * PURPOSE:
 *
 *    CRC error control for MFCC bits.
 *
 * INPUT:
 *
 *    data[] - VQ index vector, extracted from received bitstream.
 *    crc - CRC for data[], extracted from received bitstream.
 *
 * OUTPUT:
 *
 *    None.
 *
 * RETURN VALUE:
 *
 *   TRUE if data[] passes CRC check.
 *   FALSE otherwise.
 *
 *---------------------------------------------------------------------------*/


Flag mfcc_crc_errors(UWord8 data[], UWord8 crc)
{
   if (crc == mfcc_crc_encode(data)) 
   {
      return FALSE;
   }
   
   return TRUE;
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME:
 *
 *    pc_crc_encode
 *
 * PURPOSE:
 *
 *    CRC computation for pitch and class bits.
 *
 * INPUT:
 *
 *    pc_data[] - Pitch and class bits.
 *
 * OUTPUT:
 *
 *    pc_crc[] - CRC bits for pc_data.
 *
 * RETURN VALUE:
 *
 *   None.
 *
 *---------------------------------------------------------------------------*/


#ifdef XAFE
void pc_crc_encode(UWord8 pc_data[], UWord8 pc_crc[])
{

  Word16 i;

  pc_crc[0] = 0; move16();
  pc_crc[1] = 0; move16();
  for (i = 0; i < 14; i++)
  {
    logic16(); test();
    if (pc_data[i] ^ pc_crc[0])
    {
      pc_crc[0] = pc_crc[1] ^ 0x01; logic16(); move16();
      pc_crc[1] = 0x01; move16();
    }
    else
    {
      pc_crc[0] = pc_crc[1]; move16();
      pc_crc[1] = 0x00; move16();
    }
  }

  return;

}
#endif


/*----------------------------------------------------------------------------
 * FUNCTION NAME:
 *
 *    pc_crc_errors
 *
 * PURPOSE:
 *
 *    CRC error checking for pitch and class bits
 *    (bit-exact implementation not required.)
 *
 * INPUT:
 *
 *    pc_data[] - Pitch and class bits extracted from received bitstream.
 *    pc_crc[] - Pitch and class CRC bits extracted from received bitstream.
 *
 * OUTPUT:
 *
 *    None.
 *
 * RETURN VALUE:
 *
 *   TRUE if no CRC errors.
 *   FALSE otherwise.
 *
 *---------------------------------------------------------------------------*/

#ifdef XAFE
int pc_crc_errors(UWord8 pc_data[], UWord8 pc_crc[])
{
   
   UWord8 pc_crc_check[2];
   
   pc_crc_encode(pc_data,pc_crc_check);
   
   if ((pc_crc[0]^pc_crc_check[0]) || (pc_crc[1]^pc_crc_check[1]))
   {
      return TRUE;
   }
   
   return FALSE;
   
}
#endif
